# CONF

name = "lib_shopware6_api_base"
title = "python3 base API client for shopware6"
version = "v2.0.8"
url = "https://github.com/bitranox/lib_shopware6_api_base"
author = "Robert Nowotny"
author_email = "bitranox@gmail.com"
shell_command = "lib_shopware6_api_base"


def print_info() -> None:
    print(
        """\

Info for lib_shopware6_api_base:

    python3 base API client for shopware6

    Version : v2.0.8
    Url     : https://github.com/bitranox/lib_shopware6_api_base
    Author  : Robert Nowotny
    Email   : bitranox@gmail.com"""
    )
